from django.contrib.gis.db.models.fields import BaseSpatialField

from geostore.db.functions import IsEmpty as IsEmptyFunc


@BaseSpatialField.register_lookup
class IsEmpty(IsEmptyFunc):
    lookup_name = 'isempty'



from django.conf import settings

TERRA_TILES_HOSTNAMES = getattr(settings, 'TERRA_TILES_HOSTNAMES', [])

MAX_TILE_ZOOM = getattr(settings, 'MAX_TILE_ZOOM', 15)
MIN_TILE_ZOOM = getattr(settings, 'MIN_TILE_ZOOM', 10)

INTERNAL_GEOMETRY_SRID = getattr(settings, 'INTERNAL_GEOMETRY_SRID', 4326)
GEOSTORE_RELATION_CELERY_ASYNC = getattr(settings, 'GEOSTORE_RELATION_CELERY_ASYNC', False)

# LayerViewSet can be override by a subclass
GEOSTORE_LAYER_VIEWSSET = getattr(settings, 'GEOSTORE_LAYER_VIEWSSET', 'geostore.views.LayerViewSet')
# LayerSerializer can be override by a subclass
GEOSTORE_LAYER_SERIALIZER = getattr(settings, 'GEOSTORE_LAYER_SERIALIZER', 'geostore.serializers.LayerSerializer')

GEOSTORE_EXPORT_CELERY_ASYNC = getattr(settings, 'GEOSTORE_EXPORT_CELERY_ASYNC', False)



# Generated by Django 3.1.2 on 2020-10-23 12:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0042_auto_20201015_1328'),
        ('geostore', '0042_layer_routable'),
    ]

    operations = [
    ]



from django.db import transaction


def execute_async_func(async_func, args=()):
    """ Celery worker can be out of transaction, and raise DoesNotExist """
    async_func.delay(*args) if not transaction.get_connection().in_atomic_block else transaction.on_commit(
        lambda: async_func.delay(*args))



from django.utils.module_loading import import_string

from geostore import settings as app_settings
from django.urls import path, include
from rest_framework import routers

from . import views

router = routers.DefaultRouter()

# Layer ViewsSet can be override by settings
layer_viewsset = import_string(app_settings.GEOSTORE_LAYER_VIEWSSET)

router.register(r'layer', layer_viewsset)
router.register(r'group', views.LayerGroupViewsSet, basename='group'),
router.register(r'layer/(?P<layer>[\d\w\-_]+)/feature', views.FeatureViewSet,
                basename='feature')

urlpatterns = [
    path('', include(router.urls)),
]



# Generated by Django 2.0.13 on 2019-05-03 15:26

from django.db import migrations, models
import geostore.models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0022_auto_20190417_1610'),
    ]

    operations = [
        migrations.AddField(
            model_name='layer',
            name='geom_type',
            field=models.IntegerField(choices=[(geostore.GeometryTypes(0), 0), (geostore.GeometryTypes(1), 1), (geostore.GeometryTypes(3), 3), (geostore.GeometryTypes(4), 4), (geostore.GeometryTypes(5), 5), (geostore.GeometryTypes(6), 6), (geostore.GeometryTypes(7), 7)], null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-15 13:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0041_auto_20200625_1515'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='feature',
            constraint=models.CheckConstraint(check=models.Q(geom__isempty=False), name='geom_is_empty'),
        ),
    ]



from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import gettext_lazy as _


class GeostoreConfig(AppConfig):
    name = 'geostore'
    verbose_name = _("Geographic Store")

    def ready(self):
        # force use specific geojson serializer
        modules = getattr(settings, 'SERIALIZATION_MODULES', {})
        modules.update({
            'geojson': 'geostore.serializers.geojson',
        })
        setattr(settings, 'SERIALIZATION_MODULES', modules)

        import geostore.signals  # NOQA
        import geostore.db.lookups  # NOQA



# Generated by Django 3.1.3 on 2020-11-06 16:38

from django.db import migrations, models
try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
import django.db.models.deletion
import geostore.validators
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0043_merge_20201023_1209'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='identifier',
            field=models.CharField(default=uuid.uuid4, max_length=255, verbose_name='Identifier'),
        ),
        migrations.AlterField(
            model_name='feature',
            name='layer',
            field=models.ForeignKey(db_index=False, on_delete=django.db.models.deletion.PROTECT, related_name='features', to='geostore.layer', verbose_name='Layer'),
        ),
        migrations.AlterField(
            model_name='feature',
            name='properties',
            field=JSONField(blank=True, default=dict, verbose_name='Properties'),
        ),
        migrations.AlterField(
            model_name='featureextrageom',
            name='feature',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extra_geometries', to='geostore.feature', verbose_name='Feature'),
        ),
        migrations.AlterField(
            model_name='featureextrageom',
            name='identifier',
            field=models.UUIDField(blank=True, default=uuid.uuid4, editable=False, null=True, verbose_name='Identifier'),
        ),
        migrations.AlterField(
            model_name='featureextrageom',
            name='layer_extra_geom',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='features', to='geostore.layerextrageom', verbose_name='Feature'),
        ),
        migrations.AlterField(
            model_name='featureextrageom',
            name='properties',
            field=JSONField(blank=True, default=dict, verbose_name='Properties'),
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='destination',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='relations_as_destination', to='geostore.feature', verbose_name='Destination'),
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='origin',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='relations_as_origin', to='geostore.feature', verbose_name='Origin'),
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='properties',
            field=JSONField(blank=True, default=dict, verbose_name='Properties'),
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='relation',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_features', to='geostore.layerrelation', verbose_name='Relation'),
        ),
        migrations.AlterField(
            model_name='layer',
            name='authorized_groups',
            field=models.ManyToManyField(blank=True, related_name='authorized_layers', to='auth.Group', verbose_name='Authorized groups'),
        ),
        migrations.AlterField(
            model_name='layer',
            name='name',
            field=models.CharField(default=uuid.uuid4, max_length=256, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='layer',
            name='schema',
            field=JSONField(blank=True, default=dict, validators=[geostore.validators.validate_json_schema], verbose_name='Schema'),
        ),
        migrations.AlterField(
            model_name='layerextrageom',
            name='editable',
            field=models.BooleanField(default=True, verbose_name='Editable'),
        ),
        migrations.AlterField(
            model_name='layerextrageom',
            name='layer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extra_geometries', to='geostore.layer', verbose_name='Layer'),
        ),
        migrations.AlterField(
            model_name='layerextrageom',
            name='order',
            field=models.PositiveSmallIntegerField(default=0, verbose_name='Order'),
        ),
        migrations.AlterField(
            model_name='layerextrageom',
            name='title',
            field=models.CharField(max_length=250, verbose_name='Title'),
        ),
        migrations.AlterField(
            model_name='layergroup',
            name='layers',
            field=models.ManyToManyField(related_name='layer_groups', to='geostore.Layer', verbose_name='Layers'),
        ),
        migrations.AlterField(
            model_name='layergroup',
            name='name',
            field=models.CharField(max_length=256, unique=True, verbose_name='Name'),
        ),
    ]



from django.contrib.gis.serializers import geojson as basegeojson
from rest_framework_gis.serializers import GeoFeatureModelSerializer


class Serializer(basegeojson.Serializer):
    def _init_options(self):
        super()._init_options()
        self.properties_field = self.json_kwargs.pop('properties_field', None)

    def handle_field(self, obj, field):
        if field.name == self.properties_field:
            self._current = field.value_from_object(obj)
        else:
            super().handle_field(obj, field)


class FinalGeoJSONSerializer(GeoFeatureModelSerializer):
    def get_properties(self, instance, fields):
        geojson_properties = super().get_properties(instance, fields)
        instance_properties = geojson_properties.pop('properties')
        for field in instance_properties:
            geojson_properties[field] = instance_properties[field]
        return geojson_properties

    class Meta:
        geo_field = 'geom'



from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from token_tools.generator import TokenGenerator

from .models import Layer, LayerGroup


class GroupTilesToken(TokenGenerator):

    def get_groups_intersect(self, user, layergroup):
        return Group.objects.filter(
            Q(authorized_layers__in=Layer.objects.filter(layer_groups__in=[layergroup, ])) &
            Q(pk__in=user.groups.all())
        )

    def decode_idb64(self, idb64):
        decoded = super().decode_idb64(idb64)
        if decoded is not None and '-' in decoded:
            gids, lid = decoded.rsplit('-', 1)

            try:
                gids = [int(gid) for gid in gids.split('-') if gid]
                return Group.objects.filter(pk__in=gids), LayerGroup.objects.get(pk=lid)
            except (ObjectDoesNotExist, ValueError):
                pass
        return None, None

    def token_idb64(self, groups, layergroup):
        # idb64 must be group parsable content to get back all group ids
        return urlsafe_base64_encode(force_bytes('-'.join([str(g.pk) for g in groups.distinct()]) + f'-{layergroup.pk}'))

    def _make_hash_value(self, groups, layergroup):
        # We use last updated layer timestamp, so when a layer is updated token changes
        last_update = layergroup.layers.all().order_by('updated_at').first().updated_at.replace(microsecond=0, tzinfo=None)
        return '-'.join([str(g.pk) for g in groups.distinct()]) + str(last_update)


tiles_token_generator = GroupTilesToken()



# Generated by Django 3.0.2 on 2020-01-16 09:26

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0035_layerextrageom_editable'),
    ]

    operations = [
        migrations.RenameField(
            model_name='layerrelation',
            old_name='schema',
            new_name='settings',
        ),
        migrations.AddField(
            model_name='featurerelation',
            name='relation',
            field=models.ForeignKey(default='', on_delete=django.db.models.deletion.CASCADE, related_name='related_features', to='geostore.LayerRelation'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='layerrelation',
            name='exclude',
            field=JSONField(blank=True, default=dict, help_text='qs exclude (ex: {"pk__in": [...], "identifier__in":[...]}'),
        ),
        migrations.AddField(
            model_name='layerrelation',
            name='name',
            field=models.CharField(default='', max_length=250),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='layerrelation',
            name='relation_type',
            field=models.CharField(blank=True, choices=[(None, 'Manual'), ('intersects', 'Intersects'), ('distance', 'Distance')], default=(None, 'Manual'), max_length=25),
        ),
        migrations.AddField(
            model_name='layerrelation',
            name='slug',
            field=models.SlugField(default='', editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='destination',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='relations_as_destination', to='geostore.Feature'),
        ),
        migrations.AlterField(
            model_name='featurerelation',
            name='origin',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='relations_as_origin', to='geostore.Feature'),
        ),
        migrations.AlterUniqueTogether(
            name='layerrelation',
            unique_together={('name', 'origin')},
        ),
        migrations.AddConstraint(
            model_name='feature',
            constraint=models.CheckConstraint(check=models.Q(geom__isvalid=True), name='geom_is_valid'),
        ),
    ]



from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext as _


class PgRoutingMixin(models.Model):
    source = models.IntegerField(null=True,
                                 blank=True,
                                 help_text='Internal field used by pgRouting',
                                 editable=False)
    target = models.IntegerField(null=True,
                                 blank=True,
                                 help_text='Internal field used by pgRouting',
                                 editable=False)

    class Meta:
        abstract = True


class UpdateRoutingMixin(models.Model):
    routable = models.BooleanField(default=False, help_text='Used for make layer routable')

    def clean(self, *args, **kwargs):
        if self.routable and not self.is_linestring:
            raise ValidationError(_('Invalid geom type for routing'), code='invalid')
        super().clean(*args, **kwargs)

    class Meta:
        abstract = True



# Generated by Django 2.0.13 on 2019-03-26 10:49

import django.contrib.postgres.indexes
from django.db import migrations, models
from django.contrib.postgres.operations import BtreeGistExtension


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0017_auto_20181114_0806'),
    ]

    operations = [
        BtreeGistExtension(),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['identifier'], name='geo_featu_identif_212b28_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=django.contrib.postgres.indexes.GistIndex(fields=['layer', 'geom'], name='geo_featu_layer_i_3dcdde_gist'),
        ),
    ]



import json
import logging
import uuid
from itertools import islice

from django.contrib.auth.models import Group
from django.contrib.gis.db import models
from django.contrib.gis.db.models.aggregates import Extent
from django.contrib.gis.db.models.functions import Transform
from django.contrib.gis.geos import GEOSGeometry, WKBWriter
from django.contrib.gis.measure import D

from .import_export.exports import LayerExportMixin
from .import_export.imports import LayerImportMixin

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.indexes import GistIndex, GinIndex
from django.db import connection, transaction
from django.db.models import Manager
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from . import settings as app_settings
from .db.managers import FeatureQuerySet
from .db.mixins import BaseUpdatableModel, LayerBasedModelMixin
from .routing.mixins import PgRoutingMixin, UpdateRoutingMixin

from .tiles.funcs import HausdorffDistance
from .validators import (validate_geom_type, validate_json_schema,
                         validate_json_schema_data)

logger = logging.getLogger(__name__)


class Layer(LayerBasedModelMixin, LayerImportMixin, LayerExportMixin, UpdateRoutingMixin):
    name = models.CharField(max_length=256, unique=True, default=uuid.uuid4, verbose_name=_("Name"))
    schema = JSONField(default=dict, blank=True, validators=[validate_json_schema], verbose_name=_("Schema"))
    authorized_groups = models.ManyToManyField(Group, blank=True, related_name='authorized_layers',
                                               verbose_name=_("Authorized groups"))

    @transaction.atomic
    def update_geometries(self, features):
        modified = self.features.none()
        for new_feature in features:
            geometry = GEOSGeometry(json.dumps(new_feature['geometry']))
            nearest_feature = (
                self.features
                    .filter(geom__bboverlaps=geometry)  # Bounding Box Overlap
                    .annotate(hausdorff=HausdorffDistance('geom',
                                                          geometry))
                    .order_by('hausdorff')
                    .first()
            )
            feature = self.features.filter(pk=nearest_feature.pk)
            feature.update(properties=new_feature.get('properties', {}))
            modified |= feature

        return modified

    @cached_property
    def layer_properties(self):
        """
        Return properties based on layer features or layer schema definition
        """
        if self.schema:
            results = list(self.schema.get('properties', {}).keys())

        else:
            feature_table = Feature._meta.db_table

            layer_field = Feature._meta.get_field('layer').get_attname_column()[1]

            cursor = connection.cursor()
            # noinspection SqlResolve
            raw_query = f"""
                SELECT
                    jsonb_object_keys(properties) AS key
                FROM
                    (SELECT properties FROM {feature_table} WHERE {layer_field} = %s) AS t
                GROUP BY
                    key;
                """

            cursor.execute(raw_query, [self.pk, ])
            results = [x[0] for x in cursor.fetchall()]

        return {
            prop: 'str'
            for prop in results
        }

    def get_property_title(self, prop):
        """ Get json property title with its name. Return its name if not defined. """
        json_form_properties = self.schema.get('properties', {})

        if prop in json_form_properties:
            data = json_form_properties[prop]
            title = data.get('title', prop)
            return title

        return prop

    def get_property_type(self, prop):
        """ Get json property type with its name """
        prop_type = None
        json_form_properties = self.schema.get('properties', {})

        if prop in json_form_properties:
            data = json_form_properties[prop]
            prop_type = data.get('type')

        return prop_type

    def get_extent(self, srid=3857):
        return self.features.annotate(
            geom_transformed=Transform('geom', srid)
        ).aggregate(
            extent=Extent('geom_transformed')
        )

    def get_property_values(self, property_to_list):
        property_field = f'properties__{property_to_list}'

        return (
            self.features.order_by(property_field)
            .values_list(property_field, flat='true')
            .distinct(property_field)
        )

    def __str__(self):
        return f"{self.name}"

    class Meta:
        ordering = ['id']
        permissions = (
            ('can_manage_layers', 'Has all permissions on layers'),
            ('can_export_layers', 'Is able to export layers'),
            ('can_import_layers', 'Is able to import layers'),
        )


class LayerGroup(BaseUpdatableModel):
    name = models.CharField(max_length=256,
                            unique=True,
                            verbose_name=_("Name"))
    slug = models.SlugField(unique=True)
    layers = models.ManyToManyField(Layer,
                                    related_name='layer_groups',
                                    verbose_name=_("Layers"))

    def save(self, **kwargs):
        if self.pk is None:
            self.slug = slugify(self.name)
        super().save(**kwargs)


class Feature(BaseUpdatableModel, PgRoutingMixin):
    geom = models.GeometryField(srid=app_settings.INTERNAL_GEOMETRY_SRID)
    identifier = models.CharField(max_length=255,
                                  blank=False,
                                  null=False,
                                  default=uuid.uuid4,
                                  verbose_name=_("Identifier"))
    properties = JSONField(default=dict,
                           blank=True,
                           verbose_name=_("Properties"))
    layer = models.ForeignKey(Layer,
                              on_delete=models.PROTECT,
                              related_name='features',
                              db_index=False,
                              verbose_name=_("Layer"))

    objects = Manager.from_queryset(FeatureQuerySet)()

    def save(self, *args, **kwargs):
        if self.geom.hasz:
            self.geom = GEOSGeometry(WKBWriter().write(self.geom))
        super(Feature, self).save(*args, **kwargs)

    def get_bounding_box(self):
        return self.geom.extent

    def get_computed_relation_qs(self, relation):
        """ Execute relation operation to get feature queryset """
        qs_empty = Feature.objects.none()

        if relation not in self.layer.relations_as_origin.all():
            # relation should be in layer_relation_as_origins
            return qs_empty

        qs = relation.destination.features.all()
        kwargs = {}
        if relation.relation_type == 'intersects':
            kwargs.update({
                'geom__intersects': self.geom,
            })
        elif relation.relation_type == 'distance':
            kwargs.update({
                'geom__distance_lte': (self.geom,
                                       D(m=relation.settings.get('distance')),
                                       'spheroid'),
            })

        qs = qs.filter(**kwargs) if kwargs else qs
        qs = qs.exclude(**relation.exclude) if relation.exclude else qs

        return qs

    def get_stored_relation_qs(self, layer_relation):
        destination_ids = self.relations_as_origin.filter(relation=layer_relation) \
            .values_list('destination_id', flat=True)
        return Feature.objects.filter(pk__in=destination_ids)

    def sync_relations(self, layer_relation=None):
        """ replace feature relations for automatic layer relations """
        logger.info("Feature relation synchronisation")
        layer_relations = self.layer.relations_as_origin.exclude(relation_type__isnull=True)
        layer_relations = layer_relations.filter(pk__in=[layer_relation]) if layer_relation else layer_relations
        for rel in layer_relations:
            logger.info(f"relation {rel}")
            qs = self.get_computed_relation_qs(rel)
            id_list = qs.values_list("id", flat=True)
            # find relation to delete (in stored relation but not in qs result)
            to_delete = self.relations_as_origin.filter(relation=rel).exclude(destination_id__in=id_list)

            to_delete.delete()

            # find relation to add (not in stored relation but in qs
            qs = Feature.objects.filter(id__in=id_list).exclude(pk__in=self.relations_as_origin.filter(relation=rel)
                                                                .values_list('destination_id', flat=True))
            # batch creation
            batch_size = 100
            objs = (FeatureRelation(origin=self, destination=feature_rel, relation=rel) for feature_rel in qs.all())
            while True:
                batch = list(islice(objs, batch_size))
                if not batch:
                    break
                FeatureRelation.objects.bulk_create(batch, batch_size)

    @cached_property
    def relations(self):
        return {
            slugify(relation.name): self.relations_as_origin.filter(relation=relation)
            for relation in self.layer.relations_as_origin.all()
        }

    def clean(self):
        """
        Validate properties according schema if provided
        """
        validate_geom_type(self.layer.geom_type, self.geom.geom_typeid)
        validate_json_schema_data(self.properties, self.layer.schema)

    class Meta:
        ordering = ['id']
        indexes = [
            models.Index(fields=['layer', ]),
            models.Index(fields=['updated_at', ]),
            models.Index(fields=['updated_at', 'layer', ]),
            models.Index(fields=['layer', 'identifier']),
            models.Index(fields=['id', 'layer', ]),
            models.Index(fields=['source', 'layer', ]),
            models.Index(fields=['target', 'layer', ]),
            models.Index(fields=['source', 'target', 'layer']),
            GistIndex(fields=['layer', 'geom']),
            GinIndex(name='properties_gin_index', fields=['properties']),
        ]
        constraints = [
            # geometry should be valid
            models.CheckConstraint(check=models.Q(geom__isvalid=True), name='geom_is_valid'),
            # geometry should not be empty
            models.CheckConstraint(check=models.Q(geom__isempty=False), name='geom_is_empty')
        ]


class LayerRelation(models.Model):
    RELATION_TYPES = (
        (None, 'Manual'),
        ('intersects', 'Intersects'),
        ('distance', 'Distance'),
    )
    name = models.CharField(max_length=250)
    slug = models.SlugField(editable=False)
    origin = models.ForeignKey(Layer,
                               on_delete=models.PROTECT,
                               related_name='relations_as_origin')
    destination = models.ForeignKey(Layer,
                                    on_delete=models.PROTECT,
                                    related_name='relations_as_destination')
    relation_type = models.CharField(choices=RELATION_TYPES, blank=True, max_length=25, default=RELATION_TYPES[0])
    settings = JSONField(default=dict, blank=True)
    exclude = JSONField(default=dict, blank=True,
                        help_text=_("qs exclude (ex: {\"pk__in\": [...], \"identifier__in\":[...]}"))

    def save(self, *args, **kwargs):
        self.clean()
        self.slug = slugify(f'{self.origin_id}-{self.name}')
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['id']
        unique_together = (
            ('name', 'origin'),
        )


class FeatureRelation(models.Model):
    origin = models.ForeignKey(Feature,
                               on_delete=models.CASCADE,
                               related_name='relations_as_origin',
                               verbose_name=_("Origin"))
    destination = models.ForeignKey(Feature,
                                    on_delete=models.CASCADE,
                                    related_name='relations_as_destination',
                                    verbose_name=_("Destination"))
    relation = models.ForeignKey(LayerRelation,
                                 on_delete=models.CASCADE,
                                 related_name='related_features',
                                 verbose_name=_("Relation"))
    properties = JSONField(default=dict,
                           blank=True,
                           verbose_name=_("Properties"))

    class Meta:
        ordering = ['id']


class LayerExtraGeom(LayerBasedModelMixin):
    layer = models.ForeignKey(Layer,
                              on_delete=models.CASCADE,
                              related_name='extra_geometries',
                              verbose_name=_("Layer"))
    order = models.PositiveSmallIntegerField(default=0,
                                             verbose_name=_("Order"))
    slug = models.SlugField(editable=False)
    title = models.CharField(max_length=250,
                             verbose_name=_("Title"))
    editable = models.BooleanField(default=True,
                                   verbose_name=_("Editable"))

    @cached_property
    def name(self):
        return f"{slugify(self.layer.name)}-{self.slug}"

    def save(self, **kwargs):
        if self.pk is None:
            self.slug = slugify(self.title)
        super().save(**kwargs)

    def __str__(self):
        return f"{self.title} ({self.layer})"

    class Meta:
        unique_together = (
            ('layer', 'slug'),
            ('layer', 'title'),
        )
        ordering = (
            'layer', 'order'
        )


class FeatureExtraGeom(BaseUpdatableModel):
    feature = models.ForeignKey(Feature,
                                on_delete=models.CASCADE,
                                related_name='extra_geometries',
                                verbose_name=_("Feature"))
    layer_extra_geom = models.ForeignKey(LayerExtraGeom,
                                         on_delete=models.CASCADE,
                                         related_name='features',
                                         verbose_name=_("Feature"))
    geom = models.GeometryField(srid=app_settings.INTERNAL_GEOMETRY_SRID,
                                spatial_index=False)
    properties = JSONField(default=dict,
                           blank=True,
                           verbose_name=_("Properties"))
    identifier = models.UUIDField(blank=True,
                                  null=True,
                                  editable=False,
                                  default=uuid.uuid4,
                                  verbose_name=_("Identifier"))

    class Meta:
        unique_together = (
            ('feature', 'layer_extra_geom'),
        )
        indexes = [
            models.Index(fields=['layer_extra_geom', 'identifier']),
            GistIndex(name='feg_geom_gist_index', fields=['layer_extra_geom', 'geom']),
            GinIndex(name='feg_properties_gin_index', fields=['properties']),
        ]
        constraints = [
            # geometry should be valid
            models.CheckConstraint(check=models.Q(geom__isvalid=True), name='geom_extra_is_valid'),
            # geometry should not be empty
            models.CheckConstraint(check=models.Q(geom__isempty=False), name='geom_extra_is_empty')
        ]



from django.contrib.gis.db.models import FloatField, GeometryField
from django.contrib.gis.db.models.functions import GeoFunc
from django.db.models import Func


class MakeEnvelope(Func):
    function = "ST_MAKEENVELOPE"
    output_field = GeometryField()


class HausdorffDistance(GeoFunc):
    geom_param_pos = (0, 1)
    output_field = FloatField()


class SimplifyPreserveTopology(GeoFunc):
    pass


class Area(GeoFunc):
    output_field = FloatField()
